/*
 * Zmanim Java API
 * Copyright (C) 2011 - 2020 Eliyahu Hershfeld
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General
 * Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 *
 * This library is distributed in the hope that it will be useful,but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 * details.
 * You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA,
 * or connect to: http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html
 */

import 'package:intl/intl.dart';
import 'package:kosher_dart/hebrewcalendar/daf.dart';
import 'package:kosher_dart/hebrewcalendar/jewish_date.dart';
import 'package:kosher_dart/hebrewcalendar/jewish_calendar.dart';

/// The HebrewDateFormatter class formats a {@link JewishDate}.
///
/// The class formats Jewish dates in Hebrew or Latin chars, and has various settings. Sample full date output includes
/// (using various options):
/// <ul>
/// <li>21 Shevat, 5729</li>
/// <li>&#x5DB;&#x5D0; &#x5E9;&#x5D1;&#x5D8; &#x5EA;&#x5E9;&#x5DB;&#x5D8;</li>
/// <li>&#x5DB;&#x5F4;&#x5D0; &#x5E9;&#x5D1;&#x5D8; &#x5D4;&#x5F3;&#x5EA;&#x5E9;&#x5DB;&#x5F4;&#x5D8;</li>
/// <li>&#x5DB;&#x5F4;&#x5D0; &#x5E9;&#x5D1;&#x5D8; &#x5EA;&#x5E9;&#x5F4;&#x05E4; or
/// &#x5DB;&#x5F4;&#x5D0; &#x5E9;&#x5D1;&#x5D8; &#x5EA;&#x5E9;&#x5F4;&#x05E3;</li>
/// <li>&#x05DB;&#x05F3; &#x05E9;&#x05D1;&#x05D8; &#x05D5;&#x05F3; &#x05D0;&#x05DC;&#x05E4;&#x05D9;&#x05DD;</li>
/// </ul>
///
/// @see net.sourceforge.zmanim.hebrewcalendar.JewishDate
/// @see net.sourceforge.zmanim.hebrewcalendar.JewishCalendar
///
/// @author &copy; Eliyahu Hershfeld 2011 - 2020
class HebrewDateFormatter {
  bool _hebrewFormat = false;
  bool _useLonghebrewYears = false;
  bool _useGershGershayim = true;
  bool _longWeekFormat = true;
  bool _useFinalFormLetters = false;

  /// list of transliterated parshiyos using the default Ashkenazi pronounciation. The formatParsha method uses this
  /// for transliterated parsha display. This list can be overridden (for Sephardi English transliteration for example)
  /// by setting the {@link #setTransliteratedParshiosList(EnumMap)}.
  ///
  /// @see #formatParsha(JewishCalendar)
  Map<Parsha, String> _transliteratedParshaMap = {
    Parsha.NONE: "",
    Parsha.BERESHIS: "Bereshis",
    Parsha.NOACH: "Noach",
    Parsha.LECH_LECHA: "Lech Lecha",
    Parsha.VAYERA: "Vayera",
    Parsha.CHAYEI_SARA: "Chayei Sara",
    Parsha.TOLDOS: "Toldos",
    Parsha.VAYETZEI: "Vayetzei",
    Parsha.VAYISHLACH: "Vayishlach",
    Parsha.VAYESHEV: "Vayeshev",
    Parsha.MIKETZ: "Miketz",
    Parsha.VAYIGASH: "Vayigash",
    Parsha.VAYECHI: "Vayechi",
    Parsha.SHEMOS: "Shemos",
    Parsha.VAERA: "Vaera",
    Parsha.BO: "Bo",
    Parsha.BESHALACH: "Beshalach",
    Parsha.YISRO: "Yisro",
    Parsha.MISHPATIM: "Mishpatim",
    Parsha.TERUMAH: "Terumah",
    Parsha.TETZAVEH: "Tetzaveh",
    Parsha.KI_SISA: "Ki Sisa",
    Parsha.VAYAKHEL: "Vayakhel",
    Parsha.PEKUDEI: "Pekudei",
    Parsha.VAYIKRA: "Vayikra",
    Parsha.TZAV: "Tzav",
    Parsha.SHMINI: "Shmini",
    Parsha.TAZRIA: "Tazria",
    Parsha.METZORA: "Metzora",
    Parsha.ACHREI_MOS: "Achrei Mos",
    Parsha.KEDOSHIM: "Kedoshim",
    Parsha.EMOR: "Emor",
    Parsha.BEHAR: "Behar",
    Parsha.BECHUKOSAI: "Bechukosai",
    Parsha.BAMIDBAR: "Bamidbar",
    Parsha.NASSO: "Nasso",
    Parsha.BEHAALOSCHA: "Beha'aloscha",
    Parsha.SHLACH: "Sh'lach",
    Parsha.KORACH: "Korach",
    Parsha.CHUKAS: "Chukas",
    Parsha.BALAK: "Balak",
    Parsha.PINCHAS: "Pinchas",
    Parsha.MATOS: "Matos",
    Parsha.MASEI: "Masei",
    Parsha.DEVARIM: "Devarim",
    Parsha.VAESCHANAN: "Vaeschanan",
    Parsha.EIKEV: "Eikev",
    Parsha.REEH: "Re'eh",
    Parsha.SHOFTIM: "Shoftim",
    Parsha.KI_SEITZEI: "Ki Seitzei",
    Parsha.KI_SAVO: "Ki Savo",
    Parsha.NITZAVIM: "Nitzavim",
    Parsha.VAYEILECH: "Vayeilech",
    Parsha.HAAZINU: "Ha'Azinu",
    Parsha.VZOS_HABERACHA: "Vezos Habracha",
    Parsha.VAYAKHEL_PEKUDEI: "Vayakhel Pekudei",
    Parsha.TAZRIA_METZORA: "Tazria Metzora",
    Parsha.ACHREI_MOS_KEDOSHIM: "Achrei Mos Kedoshim",
    Parsha.BEHAR_BECHUKOSAI: "Behar Bechukosai",
    Parsha.CHUKAS_BALAK: "Chukas Balak",
    Parsha.MATOS_MASEI: "Matos Masei",
    Parsha.NITZAVIM_VAYEILECH: "Nitzavim Vayeilech",
    Parsha.SHKALIM: "Shekalim",
    Parsha.ZACHOR: "Zachor",
    Parsha.PARA: "Parah",
    Parsha.HACHODESH: "Hachodesh",
  };

  /// Unicode list of Hebrew parshiyos.
  Map<Parsha, String> _hebrewParshaMap = {
    Parsha.NONE: "",
    Parsha.BERESHIS: "\u05D1\u05E8\u05D0\u05E9\u05D9\u05EA",
    Parsha.NOACH: "\u05E0\u05D7",
    Parsha.LECH_LECHA: "\u05DC\u05DA \u05DC\u05DA",
    Parsha.VAYERA: "\u05D5\u05D9\u05E8\u05D0",
    Parsha.CHAYEI_SARA: "\u05D7\u05D9\u05D9 \u05E9\u05E8\u05D4",
    Parsha.TOLDOS: "\u05EA\u05D5\u05DC\u05D3\u05D5\u05EA",
    Parsha.VAYETZEI: "\u05D5\u05D9\u05E6\u05D0",
    Parsha.VAYISHLACH: "\u05D5\u05D9\u05E9\u05DC\u05D7",
    Parsha.VAYESHEV: "\u05D5\u05D9\u05E9\u05D1",
    Parsha.MIKETZ: "\u05DE\u05E7\u05E5",
    Parsha.VAYIGASH: "\u05D5\u05D9\u05D2\u05E9",
    Parsha.VAYECHI: "\u05D5\u05D9\u05D7\u05D9",
    Parsha.SHEMOS: "\u05E9\u05DE\u05D5\u05EA",
    Parsha.VAERA: "\u05D5\u05D0\u05E8\u05D0",
    Parsha.BO: "\u05D1\u05D0",
    Parsha.BESHALACH: "\u05D1\u05E9\u05DC\u05D7",
    Parsha.YISRO: "\u05D9\u05EA\u05E8\u05D5",
    Parsha.MISHPATIM: "\u05DE\u05E9\u05E4\u05D8\u05D9\u05DD",
    Parsha.TERUMAH: "\u05EA\u05E8\u05D5\u05DE\u05D4",
    Parsha.TETZAVEH: "\u05EA\u05E6\u05D5\u05D4",
    Parsha.KI_SISA: "\u05DB\u05D9 \u05EA\u05E9\u05D0",
    Parsha.VAYAKHEL: "\u05D5\u05D9\u05E7\u05D4\u05DC",
    Parsha.PEKUDEI: "\u05E4\u05E7\u05D5\u05D3\u05D9",
    Parsha.VAYIKRA: "\u05D5\u05D9\u05E7\u05E8\u05D0",
    Parsha.TZAV: "\u05E6\u05D5",
    Parsha.SHMINI: "\u05E9\u05DE\u05D9\u05E0\u05D9",
    Parsha.TAZRIA: "\u05EA\u05D6\u05E8\u05D9\u05E2",
    Parsha.METZORA: "\u05DE\u05E6\u05E8\u05E2",
    Parsha.ACHREI_MOS: "\u05D0\u05D7\u05E8\u05D9 \u05DE\u05D5\u05EA",
    Parsha.KEDOSHIM: "\u05E7\u05D3\u05D5\u05E9\u05D9\u05DD",
    Parsha.EMOR: "\u05D0\u05DE\u05D5\u05E8",
    Parsha.BEHAR: "\u05D1\u05D4\u05E8",
    Parsha.BECHUKOSAI: "\u05D1\u05D7\u05E7\u05EA\u05D9",
    Parsha.BAMIDBAR: "\u05D1\u05DE\u05D3\u05D1\u05E8",
    Parsha.NASSO: "\u05E0\u05E9\u05D0",
    Parsha.BEHAALOSCHA: "\u05D1\u05D4\u05E2\u05DC\u05EA\u05DA",
    Parsha.SHLACH: "\u05E9\u05DC\u05D7 \u05DC\u05DA",
    Parsha.KORACH: "\u05E7\u05E8\u05D7",
    Parsha.CHUKAS: "\u05D7\u05D5\u05E7\u05EA",
    Parsha.BALAK: "\u05D1\u05DC\u05E7",
    Parsha.PINCHAS: "\u05E4\u05D9\u05E0\u05D7\u05E1",
    Parsha.MATOS: "\u05DE\u05D8\u05D5\u05EA",
    Parsha.MASEI: "\u05DE\u05E1\u05E2\u05D9",
    Parsha.DEVARIM: "\u05D3\u05D1\u05E8\u05D9\u05DD",
    Parsha.VAESCHANAN: "\u05D5\u05D0\u05EA\u05D7\u05E0\u05DF",
    Parsha.EIKEV: "\u05E2\u05E7\u05D1",
    Parsha.REEH: "\u05E8\u05D0\u05D4",
    Parsha.SHOFTIM: "\u05E9\u05D5\u05E4\u05D8\u05D9\u05DD",
    Parsha.KI_SEITZEI: "\u05DB\u05D9 \u05EA\u05E6\u05D0",
    Parsha.KI_SAVO: "\u05DB\u05D9 \u05EA\u05D1\u05D5\u05D0",
    Parsha.NITZAVIM: "\u05E0\u05D9\u05E6\u05D1\u05D9\u05DD",
    Parsha.VAYEILECH: "\u05D5\u05D9\u05DC\u05DA",
    Parsha.HAAZINU: "\u05D4\u05D0\u05D6\u05D9\u05E0\u05D5",
    Parsha.VZOS_HABERACHA: "\u05D5\u05D6\u05D0\u05EA \u05D4\u05D1\u05E8\u05DB\u05D4 ",
    Parsha.VAYAKHEL_PEKUDEI: "\u05D5\u05D9\u05E7\u05D4\u05DC \u05E4\u05E7\u05D5\u05D3\u05D9",
    Parsha.TAZRIA_METZORA: "\u05EA\u05D6\u05E8\u05D9\u05E2 \u05DE\u05E6\u05E8\u05E2",
    Parsha.ACHREI_MOS_KEDOSHIM:
        "\u05D0\u05D7\u05E8\u05D9 \u05DE\u05D5\u05EA \u05E7\u05D3\u05D5\u05E9\u05D9\u05DD",
    Parsha.BEHAR_BECHUKOSAI: "\u05D1\u05D4\u05E8 \u05D1\u05D7\u05E7\u05EA\u05D9",
    Parsha.CHUKAS_BALAK: "\u05D7\u05D5\u05E7\u05EA \u05D1\u05DC\u05E7",
    Parsha.MATOS_MASEI: "\u05DE\u05D8\u05D5\u05EA \u05DE\u05E1\u05E2\u05D9",
    Parsha.NITZAVIM_VAYEILECH: "\u05E0\u05D9\u05E6\u05D1\u05D9\u05DD \u05D5\u05D9\u05DC\u05DA",
    Parsha.SHKALIM: "\u05E9\u05E7\u05DC\u05D9\u05DD",
    Parsha.ZACHOR: "\u05D6\u05DB\u05D5\u05E8",
    Parsha.PARA: "\u05E4\u05E8\u05D4",
    Parsha.HACHODESH: "\u05D4\u05D7\u05D3\u05E9",
  };

  /// Returns if the {@link #formatDayOfWeek(JewishDate)} will use the long format such as
  /// &#x05E8;&#x05D0;&#x05E9;&#x05D5;&#x05DF; or short such as &#x05D0; when formatting the day of week in
  /// {@link #isHebrewFormat() Hebrew}.
  ///
  /// @return the longWeekFormat
  /// @see #setLongWeekFormat(boolean)
  /// @see #formatDayOfWeek(JewishDate)
  bool isLongWeekFormat() {
    return _longWeekFormat;
  }

  /// Setting to control if the {@link #formatDayOfWeek(JewishDate)} will use the long format such as
  /// &#x05E8;&#x05D0;&#x05E9;&#x05D5;&#x05DF; or short such as &#x05D0; when formatting the day of week in
  /// {@link #isHebrewFormat() Hebrew}.
  ///
  /// @param longWeekFormat
  ///            the longWeekFormat to set
  void setLongWeekFormat(bool longWeekFormat) {
    this._longWeekFormat = longWeekFormat;
  }

  static const String _GERESH = "\u05F3";
  static const String _GERSHAYIM = "\u05F4";
  List<String> _transliteratedMonths = [
    "Nissan",
    "Iyar",
    "Sivan",
    "Tammuz",
    "Av",
    "Elul",
    "Tishrei",
    "Cheshvan",
    "Kislev",
    "Teves",
    "Shevat",
    "Adar",
    "Adar II",
    "Adar I"
  ];
  String _hebrewOmerPrefix = "\u05D1";
  String _transliteratedShabbosDayOfweek = "Shabbos";

  /// Returns the day of Shabbos transliterated into Latin chars. The default uses Ashkenazi pronunciation "Shabbos".
  /// This can be overwritten using the {@link #setTransliteratedShabbosDayOfWeek(String)}
  ///
  /// @return the transliteratedShabbos. The default list of months uses Ashkenazi pronunciation "Shabbos".
  /// @see #setTransliteratedShabbosDayOfWeek(String)
  /// @see #formatDayOfWeek(JewishDate)
  String getTransliteratedShabbosDayOfWeek() {
    return _transliteratedShabbosDayOfweek;
  }

  /// Setter to override the default transliterated name of "Shabbos" to alternate spelling such as "Shabbat" used by
  /// the {@link #formatDayOfWeek(JewishDate)}
  ///
  /// @param transliteratedShabbos
  ///            the transliteratedShabbos to set
  ///
  /// @see #getTransliteratedShabbosDayOfWeek()
  /// @see #formatDayOfWeek(JewishDate)
  void setTransliteratedShabbosDayOfWeek(String transliteratedShabbos) {
    this._transliteratedShabbosDayOfweek = transliteratedShabbos;
  }

  List<String> _transliteratedHolidays = [
    "Erev Pesach",
    "Pesach",
    "Chol Hamoed Pesach",
    "Pesach Sheni",
    "Erev Shavuos",
    "Shavuos",
    "Seventeenth of Tammuz",
    "Tishah B'Av",
    "Tu B'Av",
    "Erev Rosh Hashana",
    "Rosh Hashana",
    "Fast of Gedalyah",
    "Erev Yom Kippur",
    "Yom Kippur",
    "Erev Succos",
    "Succos",
    "Chol Hamoed Succos",
    "Hoshana Rabbah",
    "Shemini Atzeres",
    "Simchas Torah",
    "Erev Chanukah",
    "Chanukah",
    "Tenth of Teves",
    "Tu B'Shvat",
    "Fast of Esther",
    "Purim",
    "Shushan Purim",
    "Purim Katan",
    "Rosh Chodesh",
    "Yom HaShoah",
    "Yom Hazikaron",
    "Yom Ha'atzmaut",
    "Yom Yerushalayim"
  ];

  /// Returns the list of holidays transliterated into Latin chars. This is used by the
  /// {@link #formatYomTov(JewishCalendar)} when formatting the Yom Tov String. The default list of months uses
  /// Ashkenazi pronunciation in typical American English spelling.
  ///
  /// @return the list of holidays "Adar", "Adar II", "Adar I". The default list is currently "Erev Pesach", "Pesach",
  ///         "Chol Hamoed Pesach", "Pesach Sheni", "Erev Shavuos", "Shavuos", "Seventeenth of Tammuz", "Tishah B'Av",
  ///         "Tu B'Av", "Erev Rosh Hashana", "Rosh Hashana", "Fast of Gedalyah", "Erev Yom Kippur", "Yom Kippur",
  ///         "Erev Succos", "Succos", "Chol Hamoed Succos", "Hoshana Rabbah", "Shemini Atzeres", "Simchas Torah",
  ///         "Erev Chanukah", "Chanukah", "Tenth of Teves", "Tu B'Shvat", "Fast of Esther", "Purim", "Shushan Purim",
  ///         "Purim Katan", "Rosh Chodesh", "Yom HaShoah", "Yom Hazikaron", "Yom Ha'atzmaut", "Yom Yerushalayim"
  ///
  /// @see #setTransliteratedMonthList(String[])
  /// @see #formatYomTov(JewishCalendar)
  /// @see #isHebrewFormat()
  List<String> getTransliteratedHolidayList() {
    return _transliteratedHolidays;
  }

  /// Sets the list of holidays transliterated into Latin chars. This is used by the
  /// {@link #formatYomTov(JewishCalendar)} when formatting the Yom Tov String.
  ///
  /// @param transliteratedHolidays
  ///            the transliteratedHolidays to set. Ensure that the sequence exactly matches the list returned by the
  ///            default
  void setTransliteratedHolidayList(List<String> transliteratedHolidays) {
    this._transliteratedHolidays = _transliteratedHolidays;
  }

  /// Hebrew holiday list
  List<String> _hebrewHolidays = [
    "\u05E2\u05E8\u05D1 \u05E4\u05E1\u05D7",
    "\u05E4\u05E1\u05D7",
    "\u05D7\u05D5\u05DC \u05D4\u05DE\u05D5\u05E2\u05D3 \u05E4\u05E1\u05D7",
    "\u05E4\u05E1\u05D7 \u05E9\u05E0\u05D9",
    "\u05E2\u05E8\u05D1 \u05E9\u05D1\u05D5\u05E2\u05D5\u05EA",
    "\u05E9\u05D1\u05D5\u05E2\u05D5\u05EA",
    "\u05E9\u05D1\u05E2\u05D4 \u05E2\u05E9\u05E8 \u05D1\u05EA\u05DE\u05D5\u05D6",
    "\u05EA\u05E9\u05E2\u05D4 \u05D1\u05D0\u05D1",
    "\u05D8\u05F4\u05D5 \u05D1\u05D0\u05D1",
    "\u05E2\u05E8\u05D1 \u05E8\u05D0\u05E9 \u05D4\u05E9\u05E0\u05D4",
    "\u05E8\u05D0\u05E9 \u05D4\u05E9\u05E0\u05D4",
    "\u05E6\u05D5\u05DD \u05D2\u05D3\u05DC\u05D9\u05D4",
    "\u05E2\u05E8\u05D1 \u05D9\u05D5\u05DD \u05DB\u05D9\u05E4\u05D5\u05E8",
    "\u05D9\u05D5\u05DD \u05DB\u05D9\u05E4\u05D5\u05E8",
    "\u05E2\u05E8\u05D1 \u05E1\u05D5\u05DB\u05D5\u05EA",
    "\u05E1\u05D5\u05DB\u05D5\u05EA",
    "\u05D7\u05D5\u05DC \u05D4\u05DE\u05D5\u05E2\u05D3 \u05E1\u05D5\u05DB\u05D5\u05EA",
    "\u05D4\u05D5\u05E9\u05E2\u05E0\u05D0 \u05E8\u05D1\u05D4",
    "\u05E9\u05DE\u05D9\u05E0\u05D9 \u05E2\u05E6\u05E8\u05EA",
    "\u05E9\u05DE\u05D7\u05EA \u05EA\u05D5\u05E8\u05D4",
    "\u05E2\u05E8\u05D1 \u05D7\u05E0\u05D5\u05DB\u05D4",
    "\u05D7\u05E0\u05D5\u05DB\u05D4",
    "\u05E2\u05E9\u05E8\u05D4 \u05D1\u05D8\u05D1\u05EA",
    "\u05D8\u05F4\u05D5 \u05D1\u05E9\u05D1\u05D8",
    "\u05EA\u05E2\u05E0\u05D9\u05EA \u05D0\u05E1\u05EA\u05E8",
    "\u05E4\u05D5\u05E8\u05D9\u05DD",
    "\u05E4\u05D5\u05E8\u05D9\u05DD \u05E9\u05D5\u05E9\u05DF",
    "\u05E4\u05D5\u05E8\u05D9\u05DD \u05E7\u05D8\u05DF",
    "\u05E8\u05D0\u05E9 \u05D7\u05D5\u05D3\u05E9",
    "\u05D9\u05D5\u05DD \u05D4\u05E9\u05D5\u05D0\u05D4",
    "\u05D9\u05D5\u05DD \u05D4\u05D6\u05D9\u05DB\u05E8\u05D5\u05DF",
    "\u05D9\u05D5\u05DD \u05D4\u05E2\u05E6\u05DE\u05D0\u05D5\u05EA",
    "\u05D9\u05D5\u05DD \u05D9\u05E8\u05D5\u05E9\u05DC\u05D9\u05DD"
  ];

  /// Formats the Yom Tov (holiday) in Hebrew or transliterated Latin characters.
  ///
  /// @param jewishCalendar the JewishCalendar
  /// @return the formatted holiday or an empty String if the day is not a holiday.
  /// @see #isHebrewFormat()
  String formatYomTov(JewishCalendar jewishCalendar) {
    int index = jewishCalendar.getYomTovIndex();
    if (index == JewishCalendar.CHANUKAH) {
      int dayOfChanukah = jewishCalendar.getDayOfChanukah();
      return _hebrewFormat
          ? (formatHebrewNumber(dayOfChanukah) + " " + _hebrewHolidays[index])
          : (_transliteratedHolidays[index] + " $dayOfChanukah");
    }
    return index == -1
        ? ""
        : _hebrewFormat ? _hebrewHolidays[index] : _transliteratedHolidays[index];
  }

  /// Formats a day as Rosh Chodesh in the format of in the format of &#x05E8;&#x05D0;&#x05E9;
  /// &#x05D7;&#x05D5;&#x05D3;&#x05E9; &#x05E9;&#x05D1;&#x05D8; or Rosh Chodesh Shevat. If it
  /// is not Rosh Chodesh, an empty <code>String</code> will be returned.
  /// @param jewishCalendar the JewishCalendar
  /// @return The formatted <code>String</code> in the format of &#x05E8;&#x05D0;&#x05E9;
  /// &#x05D7;&#x05D5;&#x05D3;&#x05E9; &#x05E9;&#x05D1;&#x05D8; or Rosh Chodesh Shevat. If it
  /// is not Rosh Chodesh, an empty <code>String</code> will be returned.
  String formatRoshChodesh(JewishCalendar jewishCalendar) {
    if (!jewishCalendar.isRoshChodesh()) {
      return "";
    }
    String formattedRoshChodesh = "";
    int month = jewishCalendar.getJewishMonth();
    if (jewishCalendar.getJewishDayOfMonth() == 30) {
      if (month < JewishDate.ADAR ||
          (month == JewishDate.ADAR && jewishCalendar.isJewishLeapYear())) {
        month++;
      } else {
        // roll to Nissan
        month = JewishDate.NISSAN;
      }
    }

    // This method is only about formatting, so we shouldn't make any changes to the params passed in...
    jewishCalendar = jewishCalendar.clone() as JewishCalendar;
    jewishCalendar.setJewishMonth(month);
    formattedRoshChodesh = _hebrewFormat
        ? _hebrewHolidays[JewishCalendar.ROSH_CHODESH]
        : _transliteratedHolidays[JewishCalendar.ROSH_CHODESH];
    formattedRoshChodesh += " " + formatMonth(jewishCalendar);
    return formattedRoshChodesh;
  }

  /// Returns if the formatter is set to use Hebrew formatting in the various formatting methods.
  ///
  /// @return the hebrewFormat
  /// @see #setHebrewFormat(boolean)
  /// @see #format(JewishDate)
  /// @see #formatDayOfWeek(JewishDate)
  /// @see #formatMonth(JewishDate)
  /// @see #formatOmer(JewishCalendar)
  /// @see #formatYomTov(JewishCalendar)
  bool isHebrewFormat() {
    return _hebrewFormat;
  }

  /// Sets the formatter to format in Hebrew in the various formatting methods.
  ///
  /// @param hebrewFormat
  ///            the hebrewFormat to set
  /// @see #isHebrewFormat()
  /// @see #format(JewishDate)
  /// @see #formatDayOfWeek(JewishDate)
  /// @see #formatMonth(JewishDate)
  /// @see #formatOmer(JewishCalendar)
  /// @see #formatYomTov(JewishCalendar)
  void setHebrewFormat(bool hebrewFormat) {
    this._hebrewFormat = hebrewFormat;
  }

  /// Returns the Hebrew Omer prefix. By default it is the letter &#x5D1;, but can be set to &#x5DC; (or any other
  /// prefix) using the {@link #setHebrewOmerPrefix(String)}.
  ///
  /// @return the hebrewOmerPrefix
  ///
  /// @see #setHebrewOmerPrefix(String)
  /// @see #formatOmer(JewishCalendar)
  String getHebrewOmerPrefix() {
    return _hebrewOmerPrefix;
  }

  /// Method to set the Hebrew Omer prefix. By default it is the letter &#x5D1;, but this allows setting it to a
  /// &#x5DC; (or any other prefix).
  ///
  /// @param hebrewOmerPrefix
  ///            the hebrewOmerPrefix to set. You can use the Unicode &#92;u05DC to set it to &#x5DC;.
  /// @see #getHebrewOmerPrefix()
  /// @see #formatOmer(JewishCalendar)
  void setHebrewOmerPrefix(String hebrewOmerPrefix) {
    this._hebrewOmerPrefix = hebrewOmerPrefix;
  }

  /// Returns the list of months transliterated into Latin chars. The default list of months uses Ashkenazi
  /// pronunciation in typical American English spelling. This list has a length of 14 with 3 variations for Adar -
  /// "Adar", "Adar II", "Adar I"
  ///
  /// @return the list of months beginning in Nissan and ending in in "Adar", "Adar II", "Adar I". The default list is
  ///         currently "Nissan", "Iyar", "Sivan", "Tammuz", "Av", "Elul", "Tishrei", "Cheshvan", "Kislev", "Teves",
  ///         "Shevat", "Adar", "Adar II", "Adar I"
  /// @see #setTransliteratedMonthList(String[])
  List<String> getTransliteratedMonthList() {
    return _transliteratedMonths;
  }

  /// Setter method to allow overriding of the default list of months transliterated into into Latin chars. The default
  /// uses Ashkenazi American English transliteration.
  ///
  /// @param transliteratedMonths
  ///            an array of 14 month names such as { "Nissan", "Iyar", "Sivan", "Tamuz", "Av", "Elul", "Tishrei",
  ///            "Heshvan", "Kislev", "Tevet", "Shevat", "Adar", "Adar II", "Adar I" }
  /// @see #getTransliteratedMonthList()
  void setTransliteratedMonthList(List<String> transliteratedMonths) {
    this._transliteratedMonths = _transliteratedMonths;
  }

  /// Unicode list of Hebrew months in the format of ["\u05E0\u05D9\u05E1\u05DF","\u05D0\u05D9\u05D9\u05E8",
  /// "\u05E1\u05D9\u05D5\u05DF","\u05EA\u05DE\u05D5\u05D6","\u05D0\u05D1","\u05D0\u05DC\u05D5\u05DC",
  /// "\u05EA\u05E9\u05E8\u05D9","\u05D7\u05E9\u05D5\u05DF","\u05DB\u05E1\u05DC\u05D5","\u05D8\u05D1\u05EA",
  /// "\u05E9\u05D1\u05D8","\u05D0\u05D3\u05E8","\u05D0\u05D3\u05E8 \u05D1","\u05D0\u05D3\u05E8 \u05D0"]
  ///
  /// @see #formatMonth(JewishDate)
  static const List<String> _hebrewMonths = [
    "\u05E0\u05D9\u05E1\u05DF",
    "\u05D0\u05D9\u05D9\u05E8",
    "\u05E1\u05D9\u05D5\u05DF",
    "\u05EA\u05DE\u05D5\u05D6",
    "\u05D0\u05D1",
    "\u05D0\u05DC\u05D5\u05DC",
    "\u05EA\u05E9\u05E8\u05D9",
    "\u05D7\u05E9\u05D5\u05DF",
    "\u05DB\u05E1\u05DC\u05D5",
    "\u05D8\u05D1\u05EA",
    "\u05E9\u05D1\u05D8",
    "\u05D0\u05D3\u05E8",
    "\u05D0\u05D3\u05E8 \u05D1",
    "\u05D0\u05D3\u05E8 \u05D0"
  ];

  /// Unicode list of Hebrew days of week in the format of ["&#x05E8;&#x05D0;&#x05E9;&#x05D5;&#x05DF;",
  /// "&#x05E9;&#x05E0;&#x05D9;","&#x05E9;&#x05DC;&#x05D9;&#x05E9;&#x05D9;","&#x05E8;&#x05D1;&#x05D9;&#x05E2;&#x05D9;",
  /// "&#x05D7;&#x05DE;&#x05D9;&#x05E9;&#x05D9;","&#x05E9;&#x05E9;&#x05D9;","&#x05E9;&#x05D1;&#x05EA;"]
  static const List<String> _hebrewDaysOfWeek = [
    "\u05E8\u05D0\u05E9\u05D5\u05DF",
    "\u05E9\u05E0\u05D9",
    "\u05E9\u05DC\u05D9\u05E9\u05D9",
    "\u05E8\u05D1\u05D9\u05E2\u05D9",
    "\u05D7\u05DE\u05D9\u05E9\u05D9",
    "\u05E9\u05E9\u05D9",
    "\u05E9\u05D1\u05EA"
  ];

  /// Formats the day of week. If {@link #isHebrewFormat() Hebrew formatting} is set, it will display in the format
  /// &#x05E8;&#x05D0;&#x05E9;&#x05D5;&#x05DF; etc. If Hebrew formatting is not in use it will return it in the format
  /// of Sunday etc. There are various formatting options that will affect the output.
  ///
  /// @param jewishDate the JewishDate Object
  /// @return the formatted day of week
  /// @see #isHebrewFormat()
  /// @see #isLongWeekFormat()
  String formatDayOfWeek(JewishDate jewishDate) {
    if (_hebrewFormat) {
      if (isLongWeekFormat()) {
        return _hebrewDaysOfWeek[jewishDate.getDayOfWeek() - 1];
      } else {
        if (jewishDate.getDayOfWeek() == 7) {
          return formatHebrewNumber(300);
        } else {
          return formatHebrewNumber(jewishDate.getDayOfWeek());
        }
      }
    } else {
      if (jewishDate.getDayOfWeek() == 7) {
        if (isLongWeekFormat()) {
          return getTransliteratedShabbosDayOfWeek();
        } else {
          return getTransliteratedShabbosDayOfWeek().substring(0, 3);
        }
      } else {
        return DateFormat("EEE").format(jewishDate.getGregorianCalendar());
      }
    }
  }

  /// Returns whether the class is set to use the Geresh &#x5F3; and Gershayim &#x5F4; in formatting Hebrew dates and
  /// numbers. When true and output would look like &#x5DB;&#x5F4;&#x5D0; &#x5E9;&#x5D1;&#x5D8; &#x5EA;&#x5E9;&#x5F4;&#x5DB;
  /// (or &#x5DB;&#x5F4;&#x5D0; &#x5E9;&#x5D1;&#x5D8; &#x5EA;&#x5E9;&#x5F4;&#x5DA;). When set to false, this output
  /// would display as &#x5DB;&#x5D0; &#x5E9;&#x5D1;&#x5D8; &#x5EA;&#x5E9;&#x5DB;.
  ///
  /// @return true if set to use the Geresh &#x5F3; and Gershayim &#x5F4; in formatting Hebrew dates and numbers.
  bool isUseGershGershayim() => _useGershGershayim;

  /// Sets whether to use the Geresh &#x5F3; and Gershayim &#x5F4; in formatting Hebrew dates and numbers. The default
  /// value is true and output would look like &#x5DB;&#x5F4;&#x5D0; &#x5E9;&#x5D1;&#x5D8; &#x5EA;&#x5E9;&#x5F4;&#x5DB;
  /// (or &#x5DB;&#x5F4;&#x5D0; &#x5E9;&#x5D1;&#x5D8; &#x5EA;&#x5E9;&#x5F4;&#x5DA;). When set to false, this output would
  /// display as &#x5DB;&#x5D0; &#x5E9;&#x5D1;&#x5D8; &#x5EA;&#x5E9;&#x5DB; (or
  /// &#x5DB;&#x5D0; &#x5E9;&#x5D1;&#x5D8; &#x5EA;&#x5E9;&#x5DA;). Single digit days or month or years such as &#x05DB;&#x05F3;
  /// &#x05E9;&#x05D1;&#x05D8; &#x05D5;&#x05F3; &#x05D0;&#x05DC;&#x05E4;&#x05D9;&#x05DD; show the use of the Geresh.
  ///
  /// @param useGershGershayim
  ///            set to false to omit the Geresh &#x5F3; and Gershayim &#x5F4; in formatting
  void setUseGershGershayim(bool useGershGershayim) {
    this._useGershGershayim = useGershGershayim;
  }

  /// Returns whether the class is set to use the &#x05DE;&#x05E0;&#x05E6;&#x05E4;&#x05F4;&#x05DA; letters when
  /// formatting years ending in 20, 40, 50, 80 and 90 to produce &#x05EA;&#x05E9;&#x05F4;&#x05E4; if false or
  /// or &#x05EA;&#x05E9;&#x05F4;&#x05E3; if true. Traditionally non-final form letters are used, so the year
  /// 5780 would be formatted as &#x05EA;&#x05E9;&#x05F4;&#x05E4; if the default false is used here. If this returns
  /// true, the format &#x05EA;&#x05E9;&#x05F4;&#x05E3; would be used.
  ///
  /// @return true if set to use final form letters when formatting Hebrew years. The default value is false.
  bool isUseFinalFormLetters() => _useFinalFormLetters;

  /// When formatting a Hebrew Year, traditionally years ending in 20, 40, 50, 80 and 90 are formatted using non-final
  /// form letters for example &#x05EA;&#x05E9;&#x05F4;&#x05E4; for the year 5780. Setting this to true (the default
  /// is false) will use the final form letters for &#x05DE;&#x05E0;&#x05E6;&#x05E4;&#x05F4;&#x05DA; and will format
  /// the year 5780 as &#x05EA;&#x05E9;&#x05F4;&#x05E3;.
  ///
  /// @param useFinalFormLetters
  ///            Set this to true to use final form letters when formatting Hebrew years.
  void setUseFinalFormLetters(bool useFinalFormLetters) {
    this._useFinalFormLetters = useFinalFormLetters;
  }

  /// Returns whether the class is set to use the thousands digit when formatting. When formatting a Hebrew Year,
  /// traditionally the thousands digit is omitted and output for a year such as 5729 (1969 Gregorian) would be
  /// calculated for 729 and format as &#x5EA;&#x5E9;&#x5DB;&#x5F4;&#x5D8;. When set to true the long format year such
  /// as &#x5D4;&#x5F3; &#x5EA;&#x5E9;&#x5DB;&#x5F4;&#x5D8; for 5729/1969 is returned.
  ///
  /// @return true if set to use the thousands digit when formatting Hebrew dates and numbers.
  bool isUseLongHebrewYears() => _useLonghebrewYears;

  /// When formatting a Hebrew Year, traditionally the thousands digit is omitted and output for a year such as 5729
  /// (1969 Gregorian) would be calculated for 729 and format as &#x5EA;&#x5E9;&#x5DB;&#x5F4;&#x5D8;. This method
  /// allows setting this to true to return the long format year such as &#x5D4;&#x5F3;
  /// &#x5EA;&#x5E9;&#x5DB;&#x5F4;&#x5D8; for 5729/1969.
  ///
  /// @param useLongHebrewYears
  ///            Set this to true to use the long formatting
  void setUseLongHebrewYears(bool useLongHebrewYears) {
    this._useLonghebrewYears = useLongHebrewYears;
  }

  /// Formats the Jewish date. If the formatter is set to Hebrew, it will format in the form, "day Month year" for
  /// example &#x5DB;&#x5F4;&#x5D0; &#x5E9;&#x5D1;&#x5D8; &#x5EA;&#x5E9;&#x5DB;&#x5F4;&#x5D8;, and the format
  /// "21 Shevat, 5729" if not.
  ///
  /// @param jewishDate
  ///            the JewishDate to be formatted
  /// @return the formatted date. If the formatter is set to Hebrew, it will format in the form, "day Month year" for
  ///         example &#x5DB;&#x5F4;&#x5D0; &#x5E9;&#x5D1;&#x5D8; &#x5EA;&#x5E9;&#x5DB;&#x5F4;&#x5D8;, and the format
  ///         "21 Shevat, 5729" if not.
  String format(JewishDate jewishDate) {
    if (isHebrewFormat()) {
      return formatHebrewNumber(jewishDate.getJewishDayOfMonth()) +
          " " +
          formatMonth(jewishDate) +
          " " +
          formatHebrewNumber(jewishDate.getJewishYear());
    } else {
      return "${jewishDate.getJewishDayOfMonth()} ${formatMonth(jewishDate)}, ${jewishDate.getJewishYear()}";
    }
  }

  /// Returns a string of the current Hebrew month such as "Tishrei". Returns a string of the current Hebrew month such
  /// as "&#x5D0;&#x5D3;&#x5E8; &#x5D1;&#x5F3;".
  ///
  /// @param jewishDate
  ///            the JewishDate to format
  /// @return the formatted month name
  /// @see #isHebrewFormat()
  /// @see #setHebrewFormat(boolean)
  /// @see #getTransliteratedMonthList()
  /// @see #setTransliteratedMonthList(String[])
  String formatMonth(JewishDate jewishDate) {
    final int month = jewishDate.getJewishMonth();
    if (isHebrewFormat()) {
      if (jewishDate.isJewishLeapYear() && month == JewishDate.ADAR) {
        return _hebrewMonths[13] +
            (_useGershGershayim ? _GERESH : ""); // return Adar I, not Adar in a leap year
      } else if (jewishDate.isJewishLeapYear() && month == JewishDate.ADAR_II) {
        return _hebrewMonths[12] + (_useGershGershayim ? _GERESH : "");
      } else {
        return _hebrewMonths[month - 1];
      }
    } else {
      if (jewishDate.isJewishLeapYear() && month == JewishDate.ADAR) {
        return _transliteratedMonths[13]; // return Adar I, not Adar in a leap year
      } else {
        return _transliteratedMonths[month - 1];
      }
    }
  }

  /// Returns a String of the Omer day in the form &#x5DC;&#x5F4;&#x5D2; &#x5D1;&#x05E2;&#x05D5;&#x05DE;&#x5E8; if
  /// Hebrew Format is set, or "Omer X" or "Lag BaOmer" if not. An empty string if there is no Omer this day.
  ///
  /// @param jewishCalendar
  ///            the JewishCalendar to be formatted
  ///
  /// @return a String of the Omer day in the form or an empty string if there is no Omer this day. The default
  ///         formatting has a &#x5D1;&#x5F3; prefix that would output &#x5D1;&#x05E2;&#x05D5;&#x05DE;&#x5E8;, but this
  ///         can be set via the {@link #setHebrewOmerPrefix(String)} method to use a &#x5DC; and output
  ///         &#x5DC;&#x5F4;&#x5D2; &#x5DC;&#x05E2;&#x05D5;&#x05DE;&#x5E8;.
  /// @see #isHebrewFormat()
  /// @see #getHebrewOmerPrefix()
  /// @see #setHebrewOmerPrefix(String)
  String formatOmer(JewishCalendar jewishCalendar) {
    int omer = jewishCalendar.getDayOfOmer();
    if (omer == -1) {
      return "";
    }
    if (_hebrewFormat) {
      return formatHebrewNumber(omer) + " " + _hebrewOmerPrefix + "\u05E2\u05D5\u05DE\u05E8";
    } else {
      if (omer == 33) {
        // if lag b'omer
        return "Lag BaOmer";
      } else {
        return "Omer $omer";
      }
    }
  }

  String getLongOmer(JewishCalendar jewishCalendar) {
    int omer = jewishCalendar.getDayOfOmer();
    switch (omer) {
      case 1:
        return "הַיּוֹם יוֹם אֶחָד לָעֹמֶר:";
      case 2:
        return "הַיּוֹם שְׁנֵי יָמִים לָעֹמֶר:";
      case 3:
        return "הַיּוֹם שְׁלֹשָׁה יָמִים לָעֹמֶר:";
      case 4:
        return "הַיּוֹם אַרְבָּעָה יָמִים לָעֹמֶר:";
      case 5:
        return "הַיּוֹם חֲמִשָּׁה יָמִים לָעֹמֶר:";
      case 6:
        return "הַיּוֹם שִׁשָּׁה יָמִים לָעֹמֶר:";
      case 7:
        return "הַיּוֹם שִׁבְעָה יָמִים לָעֹמֶר, שֶׁהֵם שָׁבוּעַ אֶחָד:";
      case 8:
        return "הַיּוֹם שְׁמוֹנָה יָמִים לָעֹמֶר, שֶׁהֵם שָׁבוּעַ אֶחָד ויוֹם אֶחָד:";
      case 9:
        return "הַיּוֹם תִּשְׁעָה יָמִים לָעֹמֶר, שֶׁהֵם שָׁבוּעַ אֶחָד וּשְׁנֵי יָמִים:";
      case 10:
        return "הַיּוֹם עֲשָׂרָה יָמִים לָעֹמֶר, שֶׁהֵם שָׁבוּעַ אֶחָד וּשְׁלֹשָׁה יָמִים:";
      case 11:
        return "הַיּוֹם אַחַד עָשָׂר יוֹם לָעֹמֶר, שֶׁהֵם שָׁבוּעַ אֶחָד ואַרְבָּעָה יָמִים:";
      case 12:
        return "הַיּוֹם שְׁנֵים עָשָׂר יוֹם לָעֹמֶר, שֶׁהֵם שָׁבוּעַ אֶחָד וַחֲמִשָּׁה יָמִים:";
      case 13:
        return "הַיּוֹם שְׁלֹשָׁה עָשָׂר יוֹם לָעֹמֶר, שֶׁהֵם שָׁבוּעַ אֶחָד ושִׁשָּׁה יָמִים:";
      case 14:
        return "הַיּוֹם אַרְבָּעָה עָשָׂר יוֹם לָעֹמֶר, שֶׁהֵם שׁנֵי שָׁבוּעוֹת:";
      case 15:
        return "הַיּוֹם חֲמִשָּׁה עָשָׂר יוֹם לָעֹמֶר, שֶׁהֵם שׁנֵי שָׁבוּעוֹת ויוֹם אֶחָד:";
      case 16:
        return "הַיּוֹם שִׁשָּׁה עָשָׂר יוֹם לָעֹמֶר, שֶׁהֵם שׁנֵי שָׁבוּעוֹת וּשְׁנֵי יָמִים:";
      case 17:
        return "הַיּוֹם שִׁבְעָה עָשָׂר יוֹם לָעֹמֶר, שֶׁהֵם שׁנֵי שָׁבוּעוֹת וּשְׁלֹשָׁה יָמִים:";
      case 18:
        return "הַיּוֹם שְׁמוֹנָה עָשָׂר יוֹם לָעֹמֶר, שֶׁהֵם שׁנֵי שָׁבוּעוֹת ואַרְבָּעָה יָמִים:";
      case 19:
        return "הַיּוֹם תִּשְׁעָה עָשָׂר יוֹם לָעֹמֶר, שֶׁהֵם שׁנֵי שָׁבוּעוֹת וַחֲמִשָּׁה יָמִים:";
      case 20:
        return "הַיּוֹם עֶשְׂרִים יוֹם לָעֹמֶר, שֶׁהֵם שׁנֵי שָׁבוּעוֹת ושִׁשָּׁה יָמִים:";
      case 21:
        return "הַיּוֹם אֶחָד וְעֶשְׂרִים יוֹם לָעֹמֶר, שֶׁהֵם שׁלֹשָׁה שָׁבוּעוֹת:";
      case 22:
        return "הַיּוֹם שְׁנַיִם וְעֶשְׂרִים יוֹם לָעֹמֶר, שֶׁהֵם שׁלֹשָׁה שָׁבוּעוֹת ויוֹם אֶחָד:";
      case 23:
        return "הַיּוֹם שְׁלֹשָׁה וְעֶשְׂרִים יוֹם לָעֹמֶר, שֶׁהֵם שׁלֹשָׁה שָׁבוּעוֹת וּשְׁנֵי יָמִים:";
      case 24:
        return "הַיּוֹם אַרְבָּעָה וְעֶשְׂרִים יוֹם לָעֹמֶר, שֶׁהֵם שׁלֹשָׁה שָׁבוּעוֹת וּשְׁלֹשָׁה יָמִים:";
      case 25:
        return "הַיּוֹם חֲמִשָּׁה וְעֶשְׂרִים יוֹם לָעֹמֶר, שֶׁהֵם שׁלֹשָׁה שָׁבוּעוֹת ואַרְבָּעָה יָמִים:";
      case 26:
        return "הַיּוֹם שִׁשָּׁה וְעֶשְׂרִים יוֹם לָעֹמֶר, שֶׁהֵם שׁלֹשָׁה שָׁבוּעוֹת וַחֲמִשָּׁה יָמִים:";
      case 27:
        return "הַיּוֹם שִׁבְעָה וְעֶשְׂרִים יוֹם לָעֹמֶר, שֶׁהֵם שׁלֹשָׁה שָׁבוּעוֹת ושִׁשָּׁה יָמִים:";
      case 28:
        return "הַיּוֹם שְׁמוֹנָה וְעֶשְׂרִים יוֹם לָעֹמֶר, שֶׁהֵם אַרְבָּעָה שָׁבוּעוֹת:";
      case 29:
        return "הַיּוֹם תִּשְׁעָה וְעֶשְׂרִים יוֹם לָעֹמֶר, שֶׁהֵם אַרְבָּעָה שָׁבוּעוֹת ויוֹם אֶחָד:";
      case 30:
        return "הַיּוֹם שׁלֹשִׁים יוֹם לָעֹמֶר, שֶׁהֵם אַרְבָּעָה שָׁבוּעוֹת וּשְׁנֵי יָמִים:";
      case 31:
        return "הַיּוֹם אֶחָד וּשְׁלֹשִׁים יוֹם לָעֹמֶר, שֶׁהֵם אַרְבָּעָה שָׁבוּעוֹת וּשְׁלֹשָׁה יָמִים:";
      case 32:
        return "הַיּוֹם שְׁנַיִם וּשְׁלֹשִׁים יוֹם לָעֹמֶר, שֶׁהֵם אַרְבָּעָה שָׁבוּעוֹת ואַרְבָּעָה יָמִים:";
      case 33:
        return "הַיּוֹם שְׁלֹשָׁה וּשְׁלֹשִׁים יוֹם לָעֹמֶר, שֶׁהֵם אַרְבָּעָה שָׁבוּעוֹת וַחֲמִשָּׁה יָמִים:";
      case 34:
        return "הַיּוֹם אַרְבָּעָה וּשְׁלֹשִׁים יוֹם לָעֹמֶר, שֶׁהֵם אַרְבָּעָה שָׁבוּעוֹת ושִׁשָּׁה יָמִים:";
      case 35:
        return "הַיּוֹם חֲמִשָּׁה וּשְׁלֹשִׁים יוֹם לָעֹמֶר, שֶׁהֵם חֲמִשָּׁה שָׁבוּעוֹת:";
      case 36:
        return "הַיּוֹם שִׁשָּׁה וּשְׁלֹשִׁים יוֹם לָעֹמֶר, שֶׁהֵם חֲמִשָּׁה שָׁבוּעוֹת ויוֹם אֶחָד:";
      case 37:
        return "הַיּוֹם שִׁבְעָה וּשְׁלֹשִׁים יוֹם לָעֹמֶר, שֶׁהֵם חֲמִשָּׁה שָׁבוּעוֹת וּשְׁנֵי יָמִים:";
      case 38:
        return "הַיּוֹם שְׁמוֹנָה וּשְׁלֹשִׁים יוֹם לָעֹמֶר, שֶׁהֵם חֲמִשָּׁה שָׁבוּעוֹת וּשְׁלֹשָׁה יָמִים:";
      case 39:
        return "הַיּוֹם תִּשְׁעָה וּשְׁלֹשִׁים יוֹם לָעֹמֶר, שֶׁהֵם חֲמִשָּׁה שָׁבוּעוֹת ואַרְבָּעָה יָמִים:";
      case 40:
        return "הַיּוֹם אַרְבָּעִים יוֹם לָעֹמֶר, שֶׁהֵם חֲמִשָּׁה שָׁבוּעוֹת וַחֲמִשָּׁה יָמִים:";
      case 41:
        return "הַיּוֹם אֶחָד וְאַרְבָּעִים יוֹם לָעֹמֶר, שֶׁהֵם חֲמִשָּׁה שָׁבוּעוֹת ושִׁשָּׁה יָמִים:";
      case 42:
        return "הַיּוֹם שְׁנַיִם וְאַרְבָּעִים יוֹם לָעֹמֶר, שֶׁהֵם שִׁשָּׁה שָׁבוּעוֹת:";
      case 43:
        return "הַיּוֹם שְׁלֹשָׁה וְאַרְבָּעִים יוֹם לָעֹמֶר, שֶׁהֵם שִׁשָּׁה שָׁבוּעוֹת ויוֹם אֶחָד:";
      case 44:
        return "הַיּוֹם אַרְבָּעָה וְאַרְבָּעִים יוֹם לָעֹמֶר, שֶׁהֵם שִׁשָּׁה שָׁבוּעוֹת וּשְׁנֵי יָמִים:";
      case 45:
        return "הַיּוֹם חֲמִשָּׁה וְאַרְבָּעִים יוֹם לָעֹמֶר, שֶׁהֵם שִׁשָּׁה שָׁבוּעוֹת וּשְׁלֹשָׁה יָמִים:";
      case 46:
        return "הַיּוֹם שִׁשָּׁה וְאַרְבָּעִים יוֹם לָעֹמֶר, שֶׁהֵם שִׁשָּׁה שָׁבוּעוֹת ואַרְבָּעָה יָמִים:";
      case 47:
        return "הַיּוֹם שִׁבְעָה וְאַרְבָּעִים יוֹם לָעֹמֶר, שֶׁהֵם שִׁשָּׁה שָׁבוּעוֹת וַחֲמִשָּׁה יָמִים:";
      case 48:
        return "הַיּוֹם שְׁמוֹנָה וְאַרְבָּעִים יוֹם לָעֹמֶר, שֶׁהֵם שִׁשָּׁה שָׁבוּעוֹת ושִׁשָּׁה יָמִים:";
      case 49:
        return "הַיּוֹם תִּשְׁעָה וְאַרְבָּעִים יוֹם לָעֹמֶר, שֶׁהֵם שִׁבְעָה שָׁבוּעוֹת:";
      default:
        return "";
    }
  }

  /*
   * Experimental and incomplete
   *
   * @param moladChalakim
   * @return the formatted molad. FIXME: define proper format in English and Hebrew.
   */
  String formatMolad(double moladChalakim) {
    double adjustedChalakim = moladChalakim;
    const int MINUTE_CHALAKIM = 18;
    const int HOUR_CHALAKIM = 1080;
    const int DAY_CHALAKIM = 24 * HOUR_CHALAKIM;

    double days = adjustedChalakim / DAY_CHALAKIM;
    adjustedChalakim = adjustedChalakim - (days * DAY_CHALAKIM);
    int hours = ((adjustedChalakim ~/ HOUR_CHALAKIM));
    if (hours >= 6) {
      days += 1;
    }
    adjustedChalakim = adjustedChalakim - (hours * HOUR_CHALAKIM);
    int minutes = (adjustedChalakim ~/ MINUTE_CHALAKIM);
    adjustedChalakim = adjustedChalakim - minutes * MINUTE_CHALAKIM;
    return "Day:  ${days % 7} hours: $hours , minutes $minutes , chalakim: $adjustedChalakim";
  }

  /// Returns the kviah in the traditional 3 letter Hebrew format where the first letter represents the day of week of
  /// Rosh Hashana, the second letter represents the lengths of Cheshvan and Kislev ({@link JewishDate#SHELAIMIM
  /// Shelaimim} , {@link JewishDate#KESIDRAN Kesidran} or {@link JewishDate#CHASERIM Chaserim}) and the 3rd letter
  /// represents the day of week of Pesach. For example 5729 (1969) would return &#x5D1;&#x5E9;&#x5D4; (Rosh Hashana on
  /// Monday, Shelaimim, and Pesach on Thursday), while 5771 (2011) would return &#x5D4;&#x5E9;&#x5D2; (Rosh Hashana on
  /// Thursday, Shelaimim, and Pesach on Tuesday).
  ///
  /// @param jewishYear
  ///            the Jewish year
  /// @return the Hebrew String such as &#x5D1;&#x5E9;&#x5D4; for 5729 (1969) and &#x5D4;&#x5E9;&#x5D2; for 5771
  ///         (2011).
  String getFormattedKviah(int jewishYear) {
    JewishDate jewishDate = JewishDate.initDate(
        jewishYear: jewishYear,
        jewishMonth: JewishDate.TISHREI,
        jewishDayOfMonth: 1); // set date to Rosh Hashana
    int kviah = jewishDate.getCheshvanKislevKviah();
    int roshHashanaDayOfweek = jewishDate.getDayOfWeek();
    String returnValue = formatHebrewNumber(roshHashanaDayOfweek);
    returnValue += (kviah == JewishDate.CHASERIM
        ? "\u05D7"
        : kviah == JewishDate.SHELAIMIM ? "\u05E9" : "\u05DB");
    jewishDate.setJewishDate(jewishYear, JewishDate.NISSAN, 15); // set to Pesach of the given year
    int pesachDayOfweek = jewishDate.getDayOfWeek();
    returnValue += formatHebrewNumber(pesachDayOfweek);
    returnValue = returnValue.replaceAll(_GERESH, ""); // geresh is never used in the kviah format
    // boolean isLeapYear = JewishDate.isJewishLeapYear(jewishYear);
    // for efficiency we can avoid the expensive recalculation of the pesach day of week by adding 1 day to Rosh
    // Hashana for a 353 day year, 2 for a 354 day year, 3 for a 355 or 383 day year, 4 for a 384 day year and 5 for
    // a 385 day year
    return returnValue;
  }

  String formatDafYomiBavli(Daf daf) {
    if (_hebrewFormat) {
      return daf.getMasechta() + " " + formatHebrewNumber(daf.getDaf());
    } else {
      return daf.getMasechtaTransliterated() + " " + daf.getDaf().toString();
    }
  }

  String formatDafYomiYerushalmi(Daf daf) {
    if (_hebrewFormat) {
      String dafName = daf.getDaf() == 0 ? "" : " " + formatHebrewNumber(daf.getDaf());
      return daf.getYerushalmiMasechta() + dafName;
    } else {
      String dafName = daf.getDaf() == 0 ? "" : " ${daf.getDaf()}";
      return daf.getYerushlmiMasechtaTransliterated() + dafName;
    }
  }

  /// Returns a Hebrew formatted string of a number. The method can calculate from 0 - 9999.
  /// <ul>
  /// <li>Single digit numbers such as 3, 30 and 100 will be returned with a &#x5F3; (<a
  /// href="http://en.wikipedia.org/wiki/Geresh">Geresh</a>) appended as at the end. For example &#x5D2;&#x5F3;,
  /// &#x5DC;&#x5F3; and &#x5E7;&#x5F3;</li>
  /// <li>multi digit numbers such as 21 and 769 will be returned with a &#x5F4; (<a
  /// href="http://en.wikipedia.org/wiki/Gershayim">Gershayim</a>) between the second to last and last letters. For
  /// example &#x5DB;&#x5F4;&#x5D0;, &#x5EA;&#x5E9;&#x5DB;&#x5F4;&#x5D8;</li>
  /// <li>15 and 16 will be returned as &#x5D8;&#x5F4;&#x5D5; and &#x5D8;&#x5F4;&#x5D6;</li>
  /// <li>Single digit numbers (years assumed) such as 6000 (%1000=0) will be returned as &#x5D5;&#x5F3;
  /// &#x5D0;&#x5DC;&#x5E4;&#x5D9;&#x5DD;</li>
  /// <li>0 will return &#x5D0;&#x5E4;&#x05E1;</li>
  /// </ul>
  ///
  /// @param number
  ///            the number to be formatted. It will trow an IllegalArgumentException if the number is &lt; 0 or &gt; 9999.
  /// @return the Hebrew formatted number such as &#x5EA;&#x5E9;&#x5DB;&#x5F4;&#x5D8;
  /// @see #isUseFinalFormLetters()
  /// @see #isUseGershGershayim()
  /// @see #isHebrewFormat()
  ///
  String formatHebrewNumber(int number) {
    if (number < 0) {
      throw new ArgumentError("negative numbers can't be formatted");
    } else if (number > 9999) {
      throw new ArgumentError("numbers > 9999 can't be formatted");
    }

    const String ALAFIM = "\u05D0\u05DC\u05E4\u05D9\u05DD";
    const String EFES = "\u05D0\u05E4\u05E1";

    List<String> jHundreds = [
      "",
      "\u05E7",
      "\u05E8",
      "\u05E9",
      "\u05EA",
      "\u05EA\u05E7",
      "\u05EA\u05E8",
      "\u05EA\u05E9",
      "\u05EA\u05EA",
      "\u05EA\u05EA\u05E7"
    ];
    List<String> jTens = [
      "",
      "\u05D9",
      "\u05DB",
      "\u05DC",
      "\u05DE",
      "\u05E0",
      "\u05E1",
      "\u05E2",
      "\u05E4",
      "\u05E6"
    ];
    List<String> jTenEnds = [
      "",
      "\u05D9",
      "\u05DA",
      "\u05DC",
      "\u05DD",
      "\u05DF",
      "\u05E1",
      "\u05E2",
      "\u05E3",
      "\u05E5"
    ];
    List<String> tavTaz = ["\u05D8\u05D5", "\u05D8\u05D6"];
    List<String> jOnes = [
      "",
      "\u05D0",
      "\u05D1",
      "\u05D2",
      "\u05D3",
      "\u05D4",
      "\u05D5",
      "\u05D6",
      "\u05D7",
      "\u05D8"
    ];

    if (number == 0) {
      // do we really need this? Should it be applicable to a date?
      return EFES;
    }
    int shortNumber = number % 1000; // discard thousands
    // next check for all possible single Hebrew digit years
    bool singleDigitNumber = (shortNumber < 11 ||
        (shortNumber < 100 && shortNumber % 10 == 0) ||
        (shortNumber <= 400 && shortNumber % 100 == 0));
    int thousands = number ~/ 1000; // get # thousands
    StringBuffer sb = new StringBuffer();
    // append thousands to String
    if (number % 1000 == 0) {
      // in year is 5000, 4000 etc
      sb.write(jOnes[thousands]);
      if (isUseGershGershayim()) {
        sb.write(_GERESH);
      }
      sb.write(" ");
      sb.write(ALAFIM); // add # of thousands plus word thousand (overide alafim boolean)
      return sb.toString();
    } else if (_useLonghebrewYears && number >= 1000) {
      // if alafim boolean display thousands
      sb.write(jOnes[thousands]);
      if (isUseGershGershayim()) {
        sb.write(_GERESH); // write thousands quote
      }
      sb.write(" ");
    }
    number = number % 1000; // remove 1000s
    int hundreds = number ~/ 100; // # of hundreds
    sb.write(jHundreds[hundreds]); // add hundreds to String
    number = number % 100; // remove 100s
    if (number == 15) {
      // special case 15
      sb.write(tavTaz[0]);
    } else if (number == 16) {
      // special case 16
      sb.write(tavTaz[1]);
    } else {
      int tens = number ~/ 10;
      if (number % 10 == 0) {
        // if evenly divisable by 10
        if (singleDigitNumber == false) {
          if (isUseFinalFormLetters()) {
            sb.write(jTenEnds[tens]); // years like 5780 will end with a final form &#x05E3;
          } else {
            sb.write(jTens[tens]); // years like 5780 will end with a regular &#x05E4;
          }
        } else {
          sb.write(jTens[tens]); // standard letters so years like 5050 will end with a regular nun
        }
      } else {
        sb.write(jTens[tens]);
        number = number % 10;
        sb.write(jOnes[number]);
      }
    }
    if (isUseGershGershayim()) {
      if (singleDigitNumber == true) {
        sb.write(_GERESH); // write single quote
      } else {
        // write double quote before last digit
        String str = sb.toString();
        return '${str.substring(0, str.length - 1)}$_GERSHAYIM${str.substring(str.length - 1, str.length)}';
      }
    }
    return sb.toString();
  }

  /// Retruns the list of transliterated parshiyos used by this formatter.
  ///
  /// @return the list of transliterated Parshios
  Map<Parsha, String> getTransliteratedParshiosList() {
    return _transliteratedParshaMap;
  }

  /// Setter method to allow overriding of the default list of parshiyos transliterated into into Latin chars. The
  /// default uses Ashkenazi American English transliteration.
  ///
  /// @param transliteratedParshaMap
  ///            the transliterated Parshios as an EnumMap to set
  /// @see #getTransliteratedParshiosList()
  void setTransliteratedParshiosList(Map<Parsha, String> transliteratedParshaMap) {
    this._transliteratedParshaMap = transliteratedParshaMap;
  }

  /// Returns a String with the name of the current parsha(ios). If the formatter is set to format in Hebrew, returns
  /// a string of the current parsha(ios) in Hebrew for example &#x05D1;&#x05E8;&#x05D0;&#x05E9;&#x05D9;&#x05EA; or
  /// &#x05E0;&#x05D9;&#x05E6;&#x05D1;&#x05D9;&#x05DD; &#x05D5;&#x05D9;&#x05DC;&#x05DA; or an empty string if there
  /// are none. If not set to Hebrew, it returns a string of the parsha(ios) transliterated into Latin chars. The
  /// default uses Ashkenazi pronunciation in typical American English spelling, for example Bereshis or
  /// Nitzavim Vayeilech or an empty string if there are none.
  ///
  /// @param jewishCalendar the JewishCalendar Object
  /// @return today's parsha(ios) in Hebrew for example, if the formatter is set to format in Hebrew, returns a string
  ///         of the current parsha(ios) in Hebrew for example &#x05D1;&#x05E8;&#x05D0;&#x05E9;&#x05D9;&#x05EA; or
  ///         &#x05E0;&#x05D9;&#x05E6;&#x05D1;&#x05D9;&#x05DD; &#x05D5;&#x05D9;&#x05DC;&#x05DA; or an empty string if
  ///         there are none. If not set to Hebrew, it returns a string of the parsha(ios) transliterated into Latin
  ///         chars. The default uses Ashkenazi pronunciation in typical American English spelling, for example
  ///         Bereshis or Nitzavim Vayeilech or an empty string if there are none.
  String formatParsha(JewishCalendar jewishCalendar) {
    Parsha parsha = jewishCalendar.getParshah();
    return _hebrewFormat ? _hebrewParshaMap[parsha] : _transliteratedParshaMap[parsha];
  }

  String formatWeeklyParsha(JewishCalendar jewishCalendar) {
    int delta = 7 - jewishCalendar.getDayOfWeek();
    DateTime date = DateTime.parse(jewishCalendar.getGregorianCalendar().toIso8601String());
    return formatParsha(JewishCalendar.fromDateTime(date.add(Duration(days: delta))));
  }

  /// Returns a String with the name of the current special parsha of Shekalim, Zachor, Parah or Hachodesh or an
  /// empty String for a non-special parsha. If the formatter is set to format in Hebrew, it returns a string of
  /// the current special parsha in Hebrew, for example &#x05E9;&#x05E7;&#x05DC;&#x05D9;&#x05DD;,
  /// &#x05D6;&#x05DB;&#x05D5;&#x05E8;, &#x05E4;&#x05E8;&#x05D4; or &#x05D4;&#x05D7;&#x05D3;&#x05E9;. An empty
  /// string if the date is not a special parsha. If not set to Hebrew, it returns a string of the special parsha
  /// transliterated into Latin chars. The default uses Ashkenazi pronunciation in typical American English spelling
  /// Shekalim, Zachor, Parah or Hachodesh.
  ///
  /// @param jewishCalendar the JewishCalendar Object
  /// @return today's special parsha. If the formatter is set to format in Hebrew, returns a string
  ///         of the current special parsha  in Hebrew for in the format of &#x05E9;&#x05E7;&#x05DC;&#x05D9;&#x05DD;,
  ///         &#x05D6;&#x05DB;&#x05D5;&#x05E8;, &#x05E4;&#x05E8;&#x05D4; or &#x05D4;&#x05D7;&#x05D3;&#x05E9; or an empty
  ///         string if there are none. If not set to Hebrew, it returns a string of the special parsha transliterated
  ///         into Latin chars. The default uses Ashkenazi pronunciation in typical American English spelling of Shekalim,
  ///         Zachor, Parah or Hachodesh. An empty string if there are none.
  String formatSpecialParsha(JewishCalendar jewishCalendar) {
    Parsha specialParsha = jewishCalendar.getSpecialShabbos();
    return _hebrewFormat
        ? _hebrewParshaMap[specialParsha]
        : _transliteratedParshaMap[specialParsha];
  }
}
